#ifndef XDISTRIBUTION_H
#define XDISTRIBUTION_H

#include <string>

class observable_set;

// #include "observable.set.h"

using namespace std;

typedef enum {
    CUMULATIVE_NONE = 0,
    CUMULATIVE_LOWER,
    CUMULATIVE_UPPER,
    CUMULATIVE_BOTH
} TypeCumulative;

struct ParticleID {
  int type;
  int number;
  
  bool required_exists;
  bool required_definition;
  
  bool is_present;
  
  // number of particles of the same type
  int multiplicity;
  
  fourvector momentum;
};

class xdistribution {
 private:
  double performBinning(double value);
  void computeObservable(double &observable, double &observable_max, double &rescale_weight, double i_a);
  bool checkRequirements(int i_a);
  void fillMomenta(int i_a);
   
  observable_set *oset;
  // these are vectors of the size of the phase space oset.n_ps (i_a), which runs over different phases spaces
  // eg, for the dipoles (where the phase space may actually be differ for each of them
  vector<double> observable;     // observable to be computed in distributions
  vector<double> observable_max; // needed for cumulative distributions
  vector<vector<ParticleID> > particles;
  vector<fourvector> reconstructedParticles;
  
 public:  
  // constructors
  xdistribution();
  xdistribution(string _name, string _xdistribution_type, vector<vector<int> > _all_particle, vector<vector<int> > _all_particle_group, vector<vector<int> > _required_subparticle, double _start, double _end, int _n_bins, double _step, string _edges, string _type_binning, observable_set *_oset=NULL);
  
  void determineBin();

  vector<double> rescale_weight; // added rescale_weight to change the weight in specific distributions to compute moments
//   int xdistribution_number; 
  vector<vector<int> > all_particle;
  vector<vector<int> > all_particle_group;
  vector<vector<int> > required_subparticle;
  vector<int> required_particle;
  double start;
  double end;
  int n_bins;
  double step;
  string xdistribution_name;
  string xdistribution_type;
  int symm;

  string edges;
  string type_binning;
  int type_cumulative;
  vector<double> bin_edge;
  vector<double> bin_width;
  
  vector<int> bin;
  vector<int> bin_max;
  
  TypeCumulative typeCumulative;

  friend std::ostream & operator << ( std::ostream &, const xdistribution &);
};

#endif
